import importlib.resources

import openapi_spec_validator
import ruamel.yaml

from kodak import __about__


yaml = ruamel.yaml.YAML(typ="safe")  # pylint: disable=invalid-name


def test_openapi():
    """Validate the OpenAPI specification document structure"""
    openapi_spec_validator.validate_spec(
        yaml.load(importlib.resources.read_text("kodak", "openapi.yaml"))
    )


def test_openapi_version():
    """Check that the OpenAPI metadata matches the package metadata"""
    spec = yaml.load(importlib.resources.read_text("kodak", "openapi.yaml"))
    assert spec["info"]["version"] == __about__.__version__
    assert spec["info"]["license"]["name"] == __about__.__license__
    assert spec["info"]["title"].lower() == __about__.__title__.lower()



from kodak import database
from kodak.resources._shared import KodakResource
from kodak.resources._shared import ResponseTuple


class Heartbeat(KodakResource):
    """Expose a heartbeat endpoint to check service health"""

    routes = ("/heartbeat",)

    def get(self) -> ResponseTuple:
        """Perform a trivial database operation and return a-ok"""
        database.ImageRecord.select().count()

        return self.make_response(None)

    def head(self) -> ResponseTuple:
        """Alias HEAD to GET"""
        return self._head(self.get())



"""Shared resource base with common functionality"""
import logging
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Tuple
from typing import Union

import flask
import flask_restful

from kodak import __about__


ResponseBody = Optional[Union[Dict[str, Any], List[Dict[str, Any]], List[str]]]


ResponseHeaders = Dict[str, str]


def authenticated(func) -> Callable:
    """Decorator to wrap endpoints that need a client to authenticate to access

    .. note:: This function has no effect if ``config.private`` is set to ``False``
    """

    def _wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    # TODO: Implement this

    return _wrapper


class ResponseTuple(NamedTuple):
    """Namedtuple representing the format of a flask-restful response tuple

    :param body: Response body; must be comprised only of JSON-friendly primative types
    :param code: HTTP response code
    :param headers: Dictionary of headers
    """

    body: ResponseBody
    code: int
    headers: ResponseHeaders


class KodakResource(flask_restful.Resource):
    """Extension of the default :class:`flask_restful.Resource` class

    Add a couple of useful things to the default resource class:

    * Adds the :meth:`options` method to respond to HTTP OPTION requests
    * Adds the :meth:`_head` method as a stub helper for responding to HTTP HEAD requests
    * Adds the :meth:`make_response` method which handles response formatting boilerplate
    * Type hints the :attr:`routes` attribute for usage in subclasses
    * Adds an instance logger

    .. warning:: This class is a stub and should not be directly attached to an application

    :attribute routes: Tuple of route paths that this resource should handle; can be unpacked into
                      ``flask_restful.Api().add_route()``
    """

    routes: Tuple[str, ...]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger(__name__)

    def options(
        self, *args, **kwargs  # pylint: disable=unused-argument
    ) -> ResponseTuple:
        """Implement HTTP ``OPTIONS`` support

        `Reference documentation <https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/OPTIONS>`_
        """

        verbs = ",".join([verb.upper() for verb in flask.request.url_rule.methods])

        return self.make_response(None, 204, {"Allowed": verbs})

    def _head(self, response: ResponseTuple) -> ResponseTuple:
        """Wrapper to implement HTTP ``HEAD`` support

        `Reference documentation <https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/HEAD>`_

        .. note:: The ``head`` method cannot be implemented directly as an alias of ``get`` because
                  that would require a uniform signature for ``get`` across all resources; or some
                  hacky nonsense that wouldn't be worth it. This stub instead lets child resources
                  implement ``head`` as a oneliner.
        """
        return self.make_response(None, response.code, response.headers)

    def make_response(  # pylint: disable=no-self-use
        self,
        data: ResponseBody,
        code: int = 200,
        headers: Optional[ResponseHeaders] = None,
    ):
        """Create a response tuple from the current context

        Helper function for generating defaults, parsing common data, and formatting the response.

        :param data: Response data to return from the request
        :param code: Response code to return; defaults to `200: Ok <https://httpstatuses.com/200>`_
        :param headers: Additional headers to return with the request; the default headers will
                        be added automatically and do not need to be passed.
        :returns: Response tuple ready to be returned out of a resource method

        .. note:: This function will handle pagination and header assembly internally. The response
                  data passed to the ``data`` parameter should be unpaginated.
        """

        headers = headers or {}
        headers.update({"Server": f"{__about__.__title__}-{__about__.__version__}"})

        # 204 code specifies that it must never include a response body. Most clients will ignore
        # any response body when a 204 is given, but that's no reason to abandon best practices here
        # on the server side
        # https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/204
        return ResponseTuple(
            body=data if code != 204 else None, code=code, headers=headers
        )



"""Programatically accessible project metadata"""


__title__ = "kodak"
__version__ = "0.1.0"
__authors__ = ["Ethan Paul <24588726+enpaul@users.noreply.github.com>"]
__license__ = "MIT"
__summary__ = "HTTP server for uploading images and generating thumbnails"
__url__ = "https://github.com/enpaul/kodak/"



import logging
import shutil
from pathlib import Path
from typing import List
from typing import Optional

from kodak import configuration
from kodak import constants
from kodak import database


def identify(config: configuration.KodakConfig) -> List[database.ImageRecord]:
    """Identify source images that will be made available

    :param config: Populated application configuration object
    :returns: List of (unsaved) database models representing identified source image files
    """

    def _identify(path: Path) -> List[Path]:
        identified = []
        for item in path.iterdir():
            if item.is_file() and item.suffix in constants.IMAGE_FILE_EXTENSIONS:
                logger.debug(f"Including file {item}")
                identified.append(item.resolve())
            elif item.is_dir():
                logger.debug(f"Entering subdirectory {item}")
                identified += _identify(item)
            else:
                logger.debug(f"Skipping {item}")
        return identified

    logger = logging.getLogger(__name__)

    logger.info(
        f"Identifying image files with extensions {', '.join(constants.IMAGE_FILE_EXTENSIONS)} under {config.source_dir}"
    )

    images = _identify(config.source_dir)

    logger.info(f"Identified {len(images)} files under {config.source_dir}")

    with database.interface.atomic():
        existing = [
            item.source
            for item in database.ImageRecord.select(database.ImageRecord.source)
        ]

    logger.debug(f"Fetched {len(existing)} existing image records")

    results = []
    for image in images:
        if image.relative_to(config.source_dir) in existing:
            logger.debug(f"Skipping existing {image}")
        else:
            logger.debug(f"Including newly identified image {image}")
            results.append(database.ImageRecord.from_path(config, image))

    return results


def clean(config: configuration.KodakConfig) -> List[database.ImageRecord]:
    """Identify removed or changed source images and mark them as deleted

    :param config: Populated application configuration object
    :returns: List of (unsaved) database models representing source images that have been deleted
              or removed
    """

    logger = logging.getLogger(__name__)

    with database.interface.atomic():
        existing = database.ImageRecord.select(database.ImageRecord.source).where(
            database.ImageRecord.deleted  # pylint: disable=singleton-comparison
            == False
        )

    logger.info(f"Identified {len(existing)} existing image records")

    deleted = []
    for item in existing:
        if (config.source_dir / item.source).exists():
            logger.debug(
                f"Image file exists, record will not be modified: {item.source}"
            )
        else:
            logger.debug(f"Image file removed, record will be deleted: {item.source}")
            item.deleted = True
            deleted.append(item)

    logger.info(f"Identified {len(deleted)} image records to be marked as deleted")

    return deleted


def build(config: Optional[configuration.KodakConfig] = None) -> None:
    """Build and update the file index

    :param config: Populated application configuration object
    """
    logger = logging.getLogger(__name__)

    config = config or configuration.load()

    new_images = identify(config)
    with database.interface.atomic():
        database.ImageRecord.bulk_create(
            new_images,
            batch_size=database.calc_batch_size(config.database.backend, new_images),
        )

    removed_images = clean(config)
    with database.interface.atomic():
        database.ImageRecord.bulk_update(
            removed_images,
            fields=[database.ImageRecord.deleted],
            batch_size=database.calc_batch_size(
                config.database.backend, removed_images
            ),
        )

    logger.info(
        f"Removing generated assets for {len(removed_images)} removed image files"
    )

    for image in removed_images:
        content = config.content_dir / image.name
        logger.debug(f"Removing content directory {content}")
        shutil.rmtree(str(content))

    logger.info("Processing source links")

    with database.interface.atomic():
        for image in database.ImageRecord.select().where(
            database.ImageRecord.deleted  # pylint: disable=singleton-comparison
            == False
        ):
            if config.expose_source:
                logger.debug(f"Creating source link to {image.source}")
                image.create_link(config)
            else:
                logger.debug(f"Removing source link to {image.source}")
                image.remove_link(config)



import logging
import os
from pathlib import Path

import peewee

from kodak import configuration
from kodak import constants
from kodak.database._shared import Checksum
from kodak.database._shared import ChecksumField
from kodak.database._shared import EnumField
from kodak.database._shared import KodakModel
from kodak.database._shared import PathField


class ImageRecord(KodakModel):
    """Model for source images"""

    name = peewee.CharField(null=False)
    source = PathField(null=False)
    format_ = EnumField(constants.ImageFormat, null=False)
    deleted = peewee.BooleanField(null=False, default=False)
    checksum = ChecksumField(null=False)

    @classmethod
    def from_path(cls, config: configuration.KodakConfig, path: Path):
        """Construct an image record from a path

        :param config: Populated application configuration object
        :param path: Full path to the image file to process. The file path provided is expected to
                     already be absolute, with all symlinks and aliases resolved.
        """
        logger = logging.getLogger(__name__)

        logger.debug(f"Creating image record from path {path}")

        extension = path.suffix
        for item in constants.ImageFormat:
            if extension.replace(".", "").lower() in item.value:
                format_ = item
                logger.debug(
                    f"Identified format of file {path} as '{format_.name}' based on file extension"
                )
                break
        else:
            raise RuntimeError

        name = str(path.relative_to(config.source_dir)).replace(
            os.sep, constants.IMAGE_PATH_NAME_SEPARATOR
        )[: -len(extension)]

        logger.debug(f"Determined image name of file {path} to be '{name}'")

        return cls(
            name=name,
            source=path.relative_to(config.source_dir),
            format_=format_,
            checksum=Checksum.from_path(path),
        )

    def create_link(self, config: configuration.KodakConfig) -> Path:
        """Creates a link between the content directory and source directory

        :param config: Populated application configuration object
        :returns: Path to the created symbolic link back to the source file
        """
        logger = logging.getLogger(__name__)

        Path(config.content_dir, self.name).mkdir(exist_ok=True)
        link = Path(config.content_dir, self.name, "original")
        try:
            link.symlink_to(config.source_dir / self.source)
            logger.debug(
                f"Created link from {config.source_dir / self.source} to {link}"
            )
        except FileExistsError:
            pass
        return link

    def remove_link(self, config: configuration.KodakConfig) -> None:
        """Remove a link between the content and source directory

        :param config: Populated application configuration object
        """
        logger = logging.getLogger(__name__)
        link = Path(config.content_dir, self.name, "original")
        link.unlink(missing_ok=True)
        logger.debug(f"Removed link from {config.source_dir / self.source} to {link}")



import datetime

import flask

from kodak import database
from kodak.resources._shared import authenticated
from kodak.resources._shared import KodakResource
from kodak.resources._shared import ResponseTuple


class Image(KodakResource):
    """Handle requests for original source images"""

    routes = ("/image/<string:image_name>/original",)

    @authenticated
    def get(self, image_name: str) -> flask.Response:  # pylint: disable=no-self-use
        """Retrieve an original source image"""
        if not flask.current_app.appconfig.expose_source:
            raise RuntimeError

        with database.interface.atomic():
            image = database.ImageRecord.get(database.ImageRecord.name == image_name)

        # Note that this sends the original source file directly, rather than the symlink named
        # "original". This is because flask will serve the symlink file itself, not the linked file,
        # to the browser.
        resp = flask.send_file(
            flask.current_app.appconfig.source_dir / image.source,
            cache_timeout=int(datetime.timedelta(days=365).total_seconds()),
            add_etags=False,
        )

        resp.headers["Content-Digest"] = image.checksum.as_header()

        return resp

    def head(self, image_name: str) -> ResponseTuple:
        """Alias HEAD to GET"""
        return self._head(self.get(image_name))



import flask_restful

from kodak import index
from kodak import resources
from kodak._server import initialize_database
from kodak._server import KodakFlask
from kodak._server import make_the_tea


APPLICATION = KodakFlask(__name__)
API = flask_restful.Api(APPLICATION, catch_all_404s=True)


APPLICATION.before_request(make_the_tea)
APPLICATION.before_first_request(initialize_database)
APPLICATION.before_first_request(index.build)

for resource in resources.RESOURCES:
    API.add_resource(resource, *resource.routes)



import logging

from PIL import Image

from kodak import configuration
from kodak import constants


def scale(image: Image.Image, config: configuration.ManipConfig) -> Image.Image:
    """Scale an image to new dimensions"""

    if config.scale.strategy == constants.ScaleStrategy.ABSOLUTE:
        width = config.scale.horizontal or image.width
        height = config.scale.vertical or image.height
    elif config.scale.strategy == constants.ScaleStrategy.RELATIVE:
        width = (
            (config.scale.horizontal * image.width)
            if config.scale.vertical
            else image.width
        )
        height = (
            (config.scale.horizontal * image.height)
            if config.scale.vertical
            else image.height
        )
    else:
        raise ValueError("Here there be dragons")

    logging.getLogger(__name__).debug(
        f"Scaling image: old {image.width}x{image.height}; new {width}x{height})"
    )

    if config.scale.horizontal is None or config.scale.vertical is None:
        image = image.copy()
        image.thumbnail((width, height), Image.ANTIALIAS)
    else:
        image.resize((width, height))

    return image


def crop(image: Image.Image, config: configuration.ManipConfig) -> Image.Image:
    """Crop an image to new dimensions"""

    # TODO: add safeguards for when config values are out of bounds for the image

    width = config.crop.horizontal or image.width
    height = config.crop.vertical or image.height

    if config.crop.anchor == constants.CropAnchor.TL:
        x_1 = 0
        y_1 = 0
        x_2 = width
        y_2 = height
    elif config.crop.anchor == constants.CropAnchor.TC:
        x_1 = (image.width / 2) - (width / 2)
        y_1 = 0
        x_2 = (image.width / 2) + (width / 2)
        y_2 = height
    elif config.crop.anchor == constants.CropAnchor.TR:
        x_1 = image.width - width
        y_1 = 0
        x_2 = image.width
        y_2 = height
    elif config.crop.anchor == constants.CropAnchor.CL:
        x_1 = 0
        y_1 = (image.height / 2) - (height / 2)
        x_2 = width
        y_2 = (image.height / 2) + (height / 2)
    elif config.crop.anchor == constants.CropAnchor.C:
        x_1 = (image.width / 2) - (width / 2)
        y_1 = (image.height / 2) - (height / 2)
        x_2 = (image.width / 2) + (width / 2)
        y_2 = (image.height / 2) + (height / 2)
    elif config.crop.anchor == constants.CropAnchor.BL:
        x_1 = 0
        y_1 = image.height - height
        x_2 = width
        y_2 = image.height
    elif config.crop.anchor == constants.CropAnchor.BC:
        x_1 = (image.width / 2) - (width / 2)
        y_1 = image.height - height
        x_2 = (image.width / 2) + (width / 2)
        y_2 = image.height
    elif config.crop.anchor == constants.CropAnchor.BR:
        x_1 = image.width - width
        y_1 = image.height - height
        x_2 = image.width
        y_2 = image.height
    else:
        raise ValueError("Ye gadds! This codepath is impossible!")

    logging.getLogger(__name__).debug(
        f"Cropping image: old {image.width}x{image.height}; new {width}x{height}; upper-0 anchor ({x_1}, {y_1}); lower-image.width anchor ({x_2}, {y_2})"
    )

    return image.crop((x_1, y_1, x_2, y_2))


def black_and_white(
    image: Image.Image,
    config: configuration.ManipConfig,  # pylint: disable=unused-argument
) -> Image.Image:
    """Convert an image to full-depth black and white"""
    logger = logging.getLogger(__name__)
    logger.debug("Converting image to black and white")
    return image.convert("L")



"""Application exceptions

::

  KodakException
   +-- ClientError
   +-- ServerError
"""


class KodakException(Exception):
    """Whomp whomp, something went wrong

    But seriously, don't ever raise this exception
    """

    status: int


class ClientError(KodakException):
    """Error while processing client side input"""

    status = 400


class ImageResourceDeletedError(ClientError):
    """Requested image resource has been deleted"""

    status = 410


class IAmATeapotError(ClientError):
    """User tried to brew coffee, but application is a teapot"""

    status = 418


class ServerError(KodakException):
    """Error while processing server side data"""

    status = 500


class ImageFileRemovedError(ServerError):
    """Image file removed from server"""


class ConfigurationError(ServerError):
    """Failed to load the application configuration"""



import datetime

import flask
import peewee

from kodak import constants
from kodak import database
from kodak.resources._shared import authenticated
from kodak.resources._shared import KodakResource
from kodak.resources._shared import ResponseTuple


class ImageManip(KodakResource):
    """Handle generating and returning a processed image manip"""

    routes = ("/image/<string:image_name>/<string:manip_name>.<string:format_name>",)

    @authenticated
    def get(  # pylint: disable=no-self-use
        self, image_name: str, manip_name: str, format_name: str
    ) -> flask.Response:
        """Retrieve an image variation"""
        try:
            manip_config = flask.current_app.appconfig.manips[manip_name]
            format_ = constants.ImageFormat[format_name.upper()]
        except KeyError:
            raise RuntimeError("Manip or format doesn't exist") from None

        with database.interface.atomic():
            parent = database.ImageRecord.get(database.ImageRecord.name == image_name)

            try:
                manip = (
                    database.ManipRecord.select()
                    .where(
                        database.ManipRecord.parent == parent,
                        database.ManipRecord.name == manip_config.name,
                        database.ManipRecord.format_ == format_,
                    )
                    .get()
                )
            except peewee.DoesNotExist:
                manip = database.ManipRecord.from_parent(
                    parent, flask.current_app.appconfig, manip_config, format_
                )

        resp = flask.send_file(
            (flask.current_app.appconfig.content_dir / manip.file),
            cache_timeout=int(datetime.timedelta(days=365).total_seconds()),
            add_etags=False,
        )

        resp.headers["Content-Digest"] = manip.checksum.as_header()

        return resp

    def head(self, image_name: str, manip_name: str, format_name: str) -> ResponseTuple:
        """Alias HEAD to GET"""
        return self._head(self.get(image_name, manip_name, format_name))



from pathlib import Path

from ruamel.yaml import YAML

from kodak.resources._shared import KodakResource
from kodak.resources._shared import ResponseTuple

yaml = YAML(typ="safe")


class OpenAPI(KodakResource):
    """Handle requests for the OpenAPI specification resource"""

    routes = ("/openapi.json",)

    def get(self) -> ResponseTuple:
        """Retrieve the OpenAPI specification document"""
        with (Path(__file__).parent / "openapi.yaml").open() as infile:
            data = yaml.load(infile)

        return self.make_response(data)

    def head(self) -> ResponseTuple:
        """Alias HEAD to GET"""
        return self._head(self.get())



import logging
from typing import Sequence
from typing import Tuple
from typing import Type

import peewee

from kodak import constants
from kodak import exceptions
from kodak.configuration import KodakConfig
from kodak.database._shared import Checksum
from kodak.database._shared import INTERFACE as interface
from kodak.database._shared import KodakModel
from kodak.database.access import AccessRecord
from kodak.database.image import ImageRecord
from kodak.database.manip import ManipRecord


MODELS: Tuple[Type[KodakModel], ...] = (ImageRecord, ManipRecord, AccessRecord)


def calc_batch_size(
    backend: constants.DatabaseBackend, models: Sequence[KodakModel]
) -> int:
    """Determine the batch size that should be used when performing queries

    This is intended to work around the query variable limit in SQLite. Critically this is a
    limit to the number of _variables_, not _records_ that can be referenced in a single query.

    The "correct" way to calculate this is to iterate over the model list and tally the number of
    changed fields, then add one for the table name, and each time you reach the
    ``SQLITE_VARIABLE_LIMIT`` (which is a known constant) cut a new batch until all the models are
    processed. This is very complicated because peewee doesn't provide a simple way to reliably
    identify changed fields.

    The naive way to calculate this (i.e. the way this function does it) is to determine the
    maximum number of variables that _could be_ used to modify a record and use that as the
    constant batch limiter. The theoretical maximum number of variables associated with a single
    record is equal to the number of fields on that record, plus 1 (for the table name). This gives
    the batch size (i.e. number of records that can be modified in a single query) as:

    ::

      999 / (len(fields) + 1)

    Where ``fields`` is an array of the fields that could be written on the record.

    .. note:: This function (pretty safely) assumes that all the records in ``models`` are of the
              same model type; i.e. they all relate to the same table. This is a pretty safe
              assumption since there's no way to do multi-table updates in a single query while
              using sane SQL practices.

    .. note:: This function just returns ``len(models)`` if the backend is anything other than
              ``SQLITE``. This is because the limitation this works around is only applicable to
              SQLite so on other platforms we can just make the batch size as large as possible.

    :param backend: Backend being used by the application
    :param models: Sequence of models that need to be batched
    :returns: Number of models that can be processed in a single batch
    """
    # oof, the ratio of lines-of-docstring to lines-of-code in this function is 35:3
    if models and backend == constants.DatabaseBackend.SQLITE:
        return int(constants.SQLITE_VARIABLE_LIMIT / (len(models[0].fields) + 1))
    return len(models)


def initialize(config: KodakConfig):
    """Initialize the database interface

    Defining the database as an
    `unconfigured proxy object <http://docs.peewee-orm.com/en/latest/peewee/database.html#setting-the-database-at-run-time>`_
    allows it to be configured at runtime based on the config values.

    :param config: Populated configuration container object
    """

    logger = logging.getLogger(__name__)

    if config.database.backend == constants.DatabaseBackend.SQLITE:
        logger.debug("Using SQLite database backend")
        logger.debug(f"Applying SQLite pragmas: {config.database.sqlite.pragmas}")
        database = peewee.SqliteDatabase(
            config.database.sqlite.path, pragmas=config.database.sqlite.pragmas
        )
    elif config.database.backend == constants.DatabaseBackend.MARIADB:
        logger.debug("Using MariaDB database backend")
        logger.debug(
            "Configuring MariaDB:"
            f" {config.database.mariadb.username}@{config.database.mariadb.hostname}:{config.database.mariadb.port},"
            f" with database '{config.database.mariadb.schema}'"
        )
        database = peewee.MySQLDatabase(
            config.database.mariadb.schema,
            host=config.database.mariadb.hostname,
            port=config.database.mariadb.port,
            user=config.database.mariadb.username,
            password=config.database.mariadb.password,
            charset="utf8mb4",
        )
    else:
        raise exceptions.ConfigurationError(
            f"Invalid storage backend in configuration: {config.database.backend}"
        )

    interface.initialize(database)

    with interface.atomic():
        interface.create_tables(MODELS)



import contextlib
import json
from pathlib import Path
from typing import Dict

import pytest

from kodak import configuration
from kodak import constants
from kodak import exceptions


@contextlib.contextmanager
def mockenv(patch, env: Dict[str, str]):
    """Simple context manager for patching in a bunch of env vars"""
    for key, value in env.items():
        patch.setenv(key, value)
    yield
    for key in env:
        patch.delenv(key)


def test_conf_global(monkeypatch):
    """Test the global config object and env parser"""

    assert configuration.KodakConfig() == configuration.load()

    with mockenv(
        monkeypatch,
        {
            "KODAK_SOURCE_DIR": "foobar/baz/",
            "KODAK_CONTENT_DIR": "/var/log/run/proc/sys/class/net/dev/opt/srv",
            "KODAK_EXPOSE_SOURCE": "flalse",
            "KODAK_PRIVATE": "herblegerble",
        },
    ):
        config = configuration.load()
        assert config.source_dir == (Path.cwd() / "foobar" / "baz")
        assert config.content_dir == Path(
            "/", "var", "log", "run", "proc", "sys", "class", "net", "dev", "opt", "srv"
        )
        assert not config.expose_source
        assert not config.private

    with mockenv(
        monkeypatch, {"KODAK_EXPOSE_SOURCE": "false", "KODAK_PRIVATE": "false"}
    ):
        config = configuration.load()
        assert not config.expose_source
        assert not config.private

    with mockenv(monkeypatch, {"KODAK_EXPOSE_SOURCE": "tRuE", "KODAK_PRIVATE": "TruE"}):
        config = configuration.load()
        assert config.expose_source
        assert config.private


def test_conf_database(monkeypatch):
    """Test the database config object and env parser"""

    with mockenv(monkeypatch, {"KODAK_DATABASE_BACKEND": "couchdb"}):
        with pytest.raises(exceptions.ConfigurationError):
            configuration.load()

    pragmas = {"foo": "bar", "fizz": 1, "buzz": True}
    with mockenv(
        monkeypatch,
        {
            "KODAK_DATABASE_BACKEND": "SQLite",
            "KODAK_DATABASE_SQLITE_PRAGMAS": json.dumps(pragmas),
            "KODAK_DATABASE_SQLITE_PATH": "~/nowhere/nothing.db",
        },
    ):
        config = configuration.load()
        assert config.database.sqlite.pragmas == pragmas
        assert config.database.backend == constants.DatabaseBackend.SQLITE
        assert (
            config.database.sqlite.path
            == Path("~", "nowhere", "nothing.db").expanduser().resolve()
        )
    with mockenv(
        monkeypatch, {"KODAK_DATABASE_SQLITE_PRAGMAS": "this]is{not,valid,,js:on"}
    ):
        with pytest.raises(exceptions.ConfigurationError):
            configuration.load()

    with mockenv(
        monkeypatch,
        {
            "KODAK_DATABASE_BACKEND": "mariaDB",
        },
    ):
        assert (
            configuration.load().database.backend == constants.DatabaseBackend.MARIADB
        )
    with mockenv(monkeypatch, {"KODAK_DATABASE_MARIADB_PORT": "NaN"}):
        with pytest.raises(exceptions.ConfigurationError):
            configuration.load()


def test_conf_manip(monkeypatch):
    """Test the manipulation config object and env parser"""

    with mockenv(
        monkeypatch,
        {
            "KODAK_MANIP_BIFF_FORMATS": "png",
            "KODAK_MANIP_BIFF_BLACK_AND_WHITE": "[jonathan frakes voice] its a total fantasy",
        },
    ):
        config = configuration.load()
        assert "biff" in config.manips
        manip = config.manips["biff"]
        assert manip.name == "biff"
        assert manip.formats == {constants.ImageFormat.PNG}
        assert not manip.black_and_white

    with mockenv(
        monkeypatch,
        {
            "KODAK_MANIP_BOFF_NAME": "grand poohbah, de d*ink, of/all%of This&That",
            "KODAK_MANIP_BOFF_FORMATS": "png,jpeg,JPeg",
            "KODAK_MANIP_BOFF_BLACK_AND_WHITE": "truE",
            "KODAK_MANIP_BOFF_CROP_HORIZONTAL": "500",
            "KODAK_MANIP_BOFF_CROP_ANCHOR": "Bottom-Center",
            "KODAK_MANIP_BOFF_SCALE_VERTICAL": "1.5",
            "KODAK_MANIP_BOFF_SCALE_STRATEGY": "RelaTive",
            "KODAK_MANIP_BAFF_CROP_VERTICAL": "128",
            "KODAK_MANIP_BAFF_SCALE_HORIZONTAL": "200",
            "KODAK_MANIP_BAFF_SCALE_STRATEGY": "absoLUTE",
        },
    ):
        config = configuration.load()
        assert "grand poohbah, de d*ink, of/all%of This&That" in config.manips
        assert "baff" in config.manips

        manip = config.manips["grand poohbah, de d*ink, of/all%of This&That"]
        assert manip.name == "grand poohbah, de d*ink, of/all%of This&That"
        assert manip.formats == {constants.ImageFormat.JPEG, constants.ImageFormat.PNG}
        assert manip.black_and_white
        assert manip.crop.horizontal == 500
        assert manip.crop.anchor == constants.CropAnchor.BC
        assert manip.scale.vertical == 1.5
        assert manip.scale.strategy == constants.ScaleStrategy.RELATIVE

        manip = config.manips["baff"]
        assert manip.crop.vertical == 128
        assert manip.scale.horizontal == 200
        assert manip.scale.strategy == constants.ScaleStrategy.ABSOLUTE

    # bad format values
    with mockenv(monkeypatch, {"KODAK_MANIP_TERRIBLE_FORMATS": "jpeg,tiff,woff2"}):
        with pytest.raises(exceptions.ConfigurationError):
            configuration.load()
    # bad scale strategy value
    with mockenv(
        monkeypatch, {"KODAK_MANIP_TERRIBLE_SCALE_STRATEGY": "take it back now yall"}
    ):
        with pytest.raises(exceptions.ConfigurationError):
            configuration.load()
    # bad scale value
    with mockenv(
        monkeypatch,
        {
            "KODAK_MANIP_TERRIBLE_SCALE_STRATEGY": "absolute",
            "KODAK_MANIP_TERRIBLE_SCALE_HORIZONTAL": "32.5",
        },
    ):
        with pytest.raises(exceptions.ConfigurationError):
            configuration.load()
    # bad crop anchor value
    with mockenv(monkeypatch, {"KODAK_MANIP_TERRIBLE_CROP_ANCHOR": "ahoy"}):
        with pytest.raises(exceptions.ConfigurationError):
            configuration.load()



"""Development server stub entrypoint

Flask comes with a built-in development server. This entrypoint allows ``kodak``
to be run directly to run the development server and expose some simple config options for ease of
access. Run the below command to start the server:

::

  python -m kodak

In addition to the helpful CLI flags, the Flask development server run by this module will also
load any ``.env`` files in the current working directory when running the application.

.. warning:: As the development server will tell you on startup, do not use this for production
             deployments.
"""
import argparse
import sys

from kodak.application import APPLICATION


def main():
    """Run the development server"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-b",
        "--bind",
        help="Address or socket to bind the server to",
        default="127.0.0.1",
    )
    parser.add_argument(
        "-p", "--port", help="Port bind the server to", default=5000, type=int
    )
    parser.add_argument(
        "-D", "--debug", help="Run Flask in debug mode", action="store_true"
    )
    args = parser.parse_args()
    APPLICATION.run(host=args.bind, port=args.port, debug=args.debug, load_dotenv=True)


if __name__ == "__main__":
    sys.exit(main())



"""Test that metadata module matches pyproject"""
from pathlib import Path

import toml

from kodak import __about__


def test_about():
    """Test metadata values"""

    with (Path(__file__).resolve().parent.parent / "pyproject.toml").open() as infile:
        pyproject = toml.load(infile)

    assert pyproject["tool"]["poetry"]["name"] == __about__.__title__
    assert pyproject["tool"]["poetry"]["version"] == __about__.__version__
    assert pyproject["tool"]["poetry"]["license"] == __about__.__license__
    assert pyproject["tool"]["poetry"]["description"] == __about__.__summary__
    assert pyproject["tool"]["poetry"]["repository"] == __about__.__url__
    assert (
        all(
            item in __about__.__authors__
            for item in pyproject["tool"]["poetry"]["authors"]
        )
        is True
    )
    assert (
        all(
            item in pyproject["tool"]["poetry"]["authors"]
            for item in __about__.__authors__
        )
        is True
    )



import peewee

from kodak.database._shared import KodakModel


class AccessRecord(KodakModel):
    """Model for access keys when operating in private mode"""

    password = peewee.CharField(null=False)



import logging

import peewee
from PIL import Image

from kodak import configuration
from kodak import constants
from kodak import manipulations
from kodak.database._shared import Checksum
from kodak.database._shared import ChecksumField
from kodak.database._shared import EnumField
from kodak.database._shared import KodakModel
from kodak.database._shared import PathField
from kodak.database.image import ImageRecord


class ManipRecord(KodakModel):
    """Model for manipulated image records"""

    parent = peewee.ForeignKeyField(ImageRecord, null=False)
    name = peewee.CharField(null=False)
    file = PathField(null=False)
    format_ = EnumField(constants.ImageFormat, null=False)
    checksum = ChecksumField(null=False)

    @classmethod
    def from_parent(
        cls,
        parent: ImageRecord,
        config: configuration.KodakConfig,
        manip: configuration.ManipConfig,
        format_: constants.ImageFormat,
    ):
        """Construct an image manip record

        :param parent: Parent image record that should be manipulated
        :param config: Populated manipulation configuration object
        :param format_: Image format that the manipulation should be saved in
        :returns: Saved image manipulation record
        """
        logger = logging.getLogger(__name__)

        logger.info(
            f"Constructing manip '{manip.name}' from source file {config.source_dir / parent.source}"
        )

        filepath = (
            config.content_dir / parent.name / f"{manip.name}.{format_.name.lower()}"
        )

        with Image.open(config.source_dir / parent.source) as image:
            if manip.scale.horizontal is not None or manip.scale.vertical is not None:
                image = manipulations.scale(image, manip)

            if manip.crop.horizontal is not None or manip.crop.vertical is not None:
                image = manipulations.crop(image, manip)

            if manip.black_and_white:
                image = manipulations.black_and_white(image, manip)

            image.save(filepath, format_.name)
        logger.debug(f"Saved manipulated image at {filepath} in {format_.name} format")

        return cls(
            parent=parent,
            name=manip.name,
            file=filepath.relative_to(config.content_dir),
            checksum=Checksum.from_path(filepath),
            format_=format_,
        )



import datetime
import enum
import hashlib
import logging
import typing
import uuid
from pathlib import Path
from typing import Callable
from typing import Dict
from typing import NamedTuple
from typing import Type
from typing import Union

import peewee

if typing.TYPE_CHECKING:
    import _hashlib


INTERFACE = peewee.DatabaseProxy()


class Checksum(NamedTuple):
    """Checksum data container

    :param algorithm: Hashing algorithm, must be supported by Python's hashlib
    :param digest: Hex digest of the hash
    """

    algorithm: str
    digest: str

    @classmethod
    def from_hash(cls, data: "_hashlib.HASH"):
        """Construct from a hashlib object"""
        return cls(algorithm=data.name, digest=data.hexdigest())

    @classmethod
    def from_path(cls, path: Union[str, Path], constructor: Callable = hashlib.sha256):
        """Construct from a file path, generating the hash of the file

        .. note:: This method attempts to _efficiently_ compute a hash of large image files. The
                  hashing code was adapted from here:

                  https://stackoverflow.com/a/44873382/5361209
        """

        hasher = constructor()
        view = memoryview(bytearray(1024 * 1024))
        with Path(path).open("rb", buffering=0) as infile:
            for chunk in iter(lambda: infile.readinto(view), 0):  # type: ignore
                hasher.update(view[:chunk])

        logging.getLogger(__name__).debug(
            f"Checksum of file {path}: {hasher.name}:{hasher.hexdigest()}"
        )

        return cls.from_hash(hasher)

    def as_header(self) -> str:
        """Format the checksum for the Content-Digest HTTP header"""
        if self.algorithm.startswith("sha"):
            alg = f"sha-{self.algorithm[3:]}"
        elif self.algorithm.startswith("id"):
            alg = f"id-sha-{self.algorithm[3:]}"
        else:
            alg = self.algorithm

        return f"{alg}={self.digest}"


class EnumField(peewee.CharField):
    """Custom field for storing enums"""

    def __init__(self, enumeration: Type[enum.Enum], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.enumeration = enumeration

    def db_value(self, value: enum.Enum) -> str:
        """Convert the enum value to the database string

        :param value: Enum item to store the name of
        :raises peewee.IntegrityError: When the item passed to ``value`` is not in the field's enum
        :returns: The name of the enum item passed to ``value``
        """
        if not isinstance(value, self.enumeration):
            raise peewee.IntegrityError(
                f"Enum {self.enumeration.__name__} has no value '{value}'"
            )
        return super().db_value(value.name)

    def python_value(self, value: str) -> enum.Enum:
        """Convert the stored string to the corresponding enum

        :param value: Name of the item from the field's enum to return
        :raises peewee.IntegrityError: When the name passed to ``value`` does not correspond to an
                                       item in the field's enum
        :returns: The enum item with the name passed to ``value``
        """
        try:
            return self.enumeration[super().python_value(value)]
        except KeyError:
            raise peewee.InterfaceError(
                f"Enum {self.enumeration.__name__} has no value with name '{value}'"
            ) from None


class PathField(peewee.CharField):
    """Field for storing paths in the database"""

    def db_value(self, value: Path) -> str:
        """Serialize a pathlib object to a database string"""
        return super().db_value(str(value))

    def python_value(self, value: str) -> Path:
        """Serialize a database string to a pathlib object"""
        return Path(super().python_value(value))


class ChecksumField(peewee.CharField):
    """Field for storing checksum hashes in the database

    .. note:: The reason for implementing this is to protect against future changes to the hashing
              algorithm. Just storing the digest means that if the hashing algorithm is ever
              changed (for performance, etc) then any existing records will be invalidated. By
              storing the hashing algorithm with the digest we can protect against that possibility.
              A custom container needs to be implemented because the builtin hashlib has no way to
              recreate a hash object from the algorithm+digest without the original data.
    """

    def db_value(self, value: Checksum) -> str:
        """Serialize the checkstum to a database string"""
        return super().db_value(f"{value.algorithm}:{value.digest}")

    def python_value(self, value: str) -> Checksum:
        """Deserailize a string to a checksum container"""
        alg, _, digest = super().python_value(value).partition(":")
        return Checksum(algorithm=alg, digest=digest)


class KodakModel(peewee.Model):
    """Base model for defining common fields and attaching database"""

    class Meta:  # pylint: disable=too-few-public-methods,missing-class-docstring
        database = INTERFACE

    uuid = peewee.UUIDField(null=False, unique=True, default=uuid.uuid4)
    created = peewee.DateTimeField(null=False, default=datetime.datetime.utcnow)

    @classmethod
    @property
    def fields(cls) -> Dict[str, peewee.Field]:
        """Expose the peewee field metadata as a public object"""
        return cls._meta.fields  # pylint: disable=protected-access



import enum
from typing import Any
from typing import Dict
from typing import Set

import peewee


class DatabaseBackend(enum.Enum):
    """Enum of supported database backends"""

    MARIADB = peewee.MySQLDatabase
    SQLITE = peewee.SqliteDatabase


class ScaleStrategy(enum.Enum):
    """Available strategies for scaling images"""

    RELATIVE = enum.auto()
    ABSOLUTE = enum.auto()


class CropAnchor(enum.Enum):
    """Anchor locations for cropping images"""

    TL = "top-left"
    TC = "top-center"
    TR = "top-center"
    CL = "center-left"
    C = "center"
    CR = "center-right"
    BL = "bottom-left"
    BC = "bottom-center"
    BR = "bottom-right"


class ImageFormat(enum.Enum):
    """Supported image conversion formats"""

    JPEG = ("jpg", "jpeg")
    PNG = ("png",)


DEFAULT_SQLITE_PRAGMAS: Dict[str, Any] = {
    "journal_mode": "wal",
    "cache_size": -1 * 64000,
    "foreign_keys": 1,
    "ignore_check_constraints": 0,
    "synchronous": 0,
}

SQLITE_VARIABLE_LIMIT = 999

DEFAULT_SUPPORTED_FORMATS: Set[ImageFormat] = {ImageFormat.JPEG, ImageFormat.PNG}

IMAGE_PATH_NAME_SEPARATOR: str = "-"

IMAGE_FILE_EXTENSIONS: Set[str] = set()

for item in ImageFormat:
    for ext in item.value:
        IMAGE_FILE_EXTENSIONS.add(f".{ext}")



import flask

from kodak import configuration
from kodak import database
from kodak import exceptions


def make_the_tea() -> None:
    """Just for fun

    https://en.wikipedia.org/wiki/Hyper_Text_Coffee_Pot_Control_Protocol
    """
    if flask.request.content_type == "message/coffeepot":
        raise exceptions.IAmATeapotError(
            f"Coffee brewing request for '{flask.request.path}' cannot be completed by teapot application"
        )


def initialize_database() -> None:
    """Initialize the database connection"""
    database.initialize(flask.current_app.appconfig)


class KodakFlask(flask.Flask):
    """Extend the default Flask object to add the custom application config

    There's probably an easier/more kosher way to do this, but ¯\\_(ツ)_/¯
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.appconfig: configuration.KodakConfig = configuration.load()



from typing import Tuple
from typing import Type

from kodak.resources._shared import KodakResource
from kodak.resources.heartbeat import Heartbeat
from kodak.resources.image import Image
from kodak.resources.manip import ImageManip
from kodak.resources.openapi import OpenAPI


RESOURCES: Tuple[Type[KodakResource], ...] = (
    Heartbeat,
    Image,
    ImageManip,
    OpenAPI,
)



import enum
import json
import os
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Optional
from typing import Set
from typing import Type
from typing import Union

from kodak import constants
from kodak import exceptions


def _get_int(var: str, default: Optional[int]) -> Optional[int]:
    return int(os.environ[var]) if var in os.environ else default


def _get_float(var: str, default: Optional[float]) -> Optional[float]:
    return float(os.environ[var]) if var in os.environ else default


def _get_enum_by_name(
    var: str, enumeration: Type[enum.Enum], default: enum.Enum
) -> enum.Enum:
    return enumeration[os.environ[var].upper()] if var in os.environ else default


def _get_enum_by_value(
    var: str, enumeration: Type[enum.Enum], default: enum.Enum
) -> enum.Enum:
    return enumeration(os.environ[var].lower()) if var in os.environ else default


def _get_path(var: str, default: Union[str, Path]) -> Path:
    return Path(os.environ.get(var, default)).expanduser().resolve()


def _get_bool(var: str, default: bool) -> bool:
    return os.getenv(var, str(default)).lower() == "true"


@dataclass
class DatabaseSqliteConfig:
    """SQLite database backend configuration options

    :param path: Path to the SQLite database file
    :param pragmas: Mapping of SQLite pragmas to apply to the database connection
    """

    path: Path = Path.cwd() / "kodak.db"
    pragmas: Dict[str, Any] = field(
        default_factory=lambda: constants.DEFAULT_SQLITE_PRAGMAS
    )

    @classmethod
    def from_env(cls):
        """Build dataclass from environment"""
        return cls(
            path=_get_path("KODAK_DATABASE_SQLITE_PATH", cls.path),
            pragmas=json.loads(os.environ["KODAK_DATABASE_SQLITE_PRAGMAS"])
            if "KODAK_DATABASE_SQLITE_PRAGMAS" in os.environ
            else constants.DEFAULT_SQLITE_PRAGMAS,
        )


@dataclass
class DatabaseMariaConfig:
    """MariaDB database backend configuration options

    :param hostname: Hostname or IP address of the host running the database server
    :param username: Username of the account to use for connecting to the database server
    :param password: Password for the account to use for connecting to the database server
    :param port: Port on the host that the database server is listening on
    :param schema: Database schema that the application should use
    """

    hostname: str = "localhost"
    username: str = "root"
    password: Optional[str] = None
    port: int = 3306
    schema: str = "kodak"

    @classmethod
    def from_env(cls):
        """Build dataclass from environment"""
        return cls(
            hostname=os.getenv("KODAK_DATABASE_MARIADB_HOSTNAME", cls.hostname),
            username=os.getenv("KODAK_DATABASE_MARIADB_USERNAME", cls.username),
            password=os.environ.get("KODAK_DATABASE_MARIADB_PASSWORD", cls.password),
            port=_get_int("KODAK_DATABASE_MARIADB_PORT", cls.port),
            schema=os.getenv("KODAK_DATABASE_MARIADB_SCHEMA", cls.schema),
        )


@dataclass
class DatabaseConfig:
    """Database backend configuration

    :param backend: Enum selecting the backend to use for storing data
    :param sqlite: Container of SQLite settings
    :param mariadb: Container of MariaDB settings
    """

    backend: constants.DatabaseBackend = constants.DatabaseBackend.SQLITE
    sqlite: DatabaseSqliteConfig = field(default_factory=DatabaseSqliteConfig.from_env)
    mariadb: DatabaseMariaConfig = field(default_factory=DatabaseMariaConfig.from_env)

    @classmethod
    def from_env(cls):
        """Build dataclass from environment"""
        return cls(
            backend=_get_enum_by_name(
                "KODAK_DATABASE_BACKEND", constants.DatabaseBackend, cls.backend
            )
        )


@dataclass
class ManipCropConfig:
    """Settings for cropping an image

    :param horizontal: Size the image should be cropped to (in pixels) in the horizontal direction
    :param vertical: Size the image should be cropped to (in pixels) in the vertical direction
    :param anchor: Image location anchor that cropping should be done relative to
    """

    horizontal: Optional[int] = None
    vertical: Optional[int] = None
    anchor: constants.CropAnchor = constants.CropAnchor.C

    @classmethod
    def from_env(cls, key: str):
        """Build dataclass from environment"""
        return cls(
            anchor=_get_enum_by_value(  # type: ignore
                f"KODAK_MANIP_{key}_CROP_ANCHOR", constants.CropAnchor, cls.anchor
            ),
            horizontal=_get_int(f"KODAK_MANIP_{key}_CROP_HORIZONTAL", cls.horizontal),
            vertical=_get_int(f"KODAK_MANIP_{key}_CROP_VERTICAL", cls.vertical),
        )


@dataclass
class ManipScaleConfig:
    """Settings for scaling an image

    :param horizontal: Horizontal scaling dimension. If ``strategy`` is ``ABSOLUTE`` then this is
                       the pixel measurement that the horizontal dimension will be scaled up or
                       down to; if ``strategy`` is ``RELATIVE`` then this is a percentage modifier
                       that will be applied to the image's existing horizontal dimension.
    :param vertical: Vertical scaling dimension. If ``strategy`` is ``ABSOLUTE`` then this is the
                     pixel measurement that the vertical dimension will be scaled up or down to;
                     if ``strategy`` is ``RELATIVE`` then this is a percentage modifier that will
                     be applied to the image's existing vertical dimension.
    :param strategy: Strategy to use for scaling the image. Use ``ABSOLUTE`` to scale to an
                     absolute pixel measurement and use ``RELATIVE`` to scale relative to the
                     existing dimensions.
    """

    horizontal: Optional[Union[int, float]] = None
    vertical: Optional[Union[int, float]] = None
    strategy: constants.ScaleStrategy = constants.ScaleStrategy.ABSOLUTE

    @classmethod
    def from_env(cls, key: str):
        """Build dataclass from environment"""
        strategy = _get_enum_by_name(
            f"KODAK_MANIP_{key}_SCALE_STRATEGY", constants.ScaleStrategy, cls.strategy
        )

        if strategy == constants.ScaleStrategy.ABSOLUTE:
            parser = _get_int  # type: ignore
        elif strategy == constants.ScaleStrategy.RELATIVE:
            parser = _get_float  # type: ignore
        else:
            raise RuntimeError("This path should not be possible")

        return cls(
            strategy=strategy,  # type: ignore
            vertical=parser(f"KODAK_MANIP_{key}_SCALE_VERTICAL", cls.vertical),  # type: ignore
            horizontal=parser(f"KODAK_MANIP_{key}_SCALE_HORIZONTAL", cls.horizontal),  # type: ignore
        )


@dataclass
class ManipConfig:
    """Image manipulation configuration settings

    :param name: Name of the manipulation that will be accessed in the URL
    :param crop: Contaienr of settings for cropping an image
    :param scale: Container of settings for scaling an image
    :param formats: Set of image formats that the source can be dynamically converted into
    :param black_and_white: Whether the image should be converted to black and white
    """

    name: str
    crop: ManipCropConfig = field(default_factory=ManipCropConfig)
    scale: ManipScaleConfig = field(default_factory=ManipScaleConfig)
    formats: Set[constants.ImageFormat] = field(
        default_factory=lambda: constants.DEFAULT_SUPPORTED_FORMATS
    )
    black_and_white: bool = False

    # TODO: Implement support for these settings
    # brightness: int = 0
    # contrast: int = 0
    # sepia: bool = False

    @classmethod
    def from_env(cls, key: str):
        """Build dataclass from environment"""
        name = os.getenv(f"KODAK_MANIP_{key}_NAME", key.lower())

        if name == "original":
            raise exceptions.ConfigurationError(
                "Manipulation name 'original' is reserved for application usage"
            )

        return cls(
            name=name,
            crop=ManipCropConfig.from_env(key),
            scale=ManipScaleConfig.from_env(key),
            formats=set(
                constants.ImageFormat[item.strip().upper()]
                for item in os.environ[f"KODAK_MANIP_{key}_FORMATS"].split(",")
            )
            if f"KODAK_MANIP_{key}_FORMATS" in os.environ
            else constants.DEFAULT_SUPPORTED_FORMATS,
            black_and_white=_get_bool(
                f"KODAK_MANIP_{key}_BLACK_AND_WHITE", cls.black_and_white
            ),
        )


@dataclass
class KodakConfig:
    """Global application configuration settings

    :param database: Container of database backend settings
    :param manips: Mapping of manipulation config names to image manipulation configurations
    :param source_dir: Path to where source images should be loaded from
    :param content_dir: Path to where the application should store generated images
    :param expose_source: Whether the original image should be exposed to clients
    :param private: Whether authentication is required for accessing the server
    """

    database: DatabaseConfig = field(default_factory=DatabaseConfig.from_env)
    manips: Dict[str, ManipConfig] = field(default_factory=dict)
    source_dir: Path = Path.cwd() / "pictures"
    content_dir: Path = Path.cwd() / "content"
    expose_source: bool = False
    private: bool = False

    @classmethod
    def from_env(cls):
        """Build dataclass from environment"""
        manip_names = set(
            key.replace("KODAK_MANIP_", "").partition("_")[0]
            for key in os.environ
            if key.startswith("KODAK_MANIP_")
        )
        manips = [ManipConfig.from_env(name) for name in manip_names]
        return cls(
            source_dir=_get_path("KODAK_SOURCE_DIR", cls.source_dir),
            content_dir=_get_path("KODAK_CONTENT_DIR", cls.content_dir),
            expose_source=_get_bool("KODAK_EXPOSE_SOURCE", cls.expose_source),
            private=_get_bool("KODAK_PRIVATE", cls.private),
            manips={item.name: item for item in manips},
        )


def load() -> KodakConfig:
    """Load the application configuration from environment variables

    :returns: Populated environment configuration
    """
    try:
        return KodakConfig.from_env()
    except (ValueError, TypeError, IndexError, KeyError) as err:
        raise exceptions.ConfigurationError(f"Failed to load configuration: {err}")



import argparse
import logging
import sys

from kodak import __about__
from kodak import configuration
from kodak import database
from kodak import index


def get_args() -> argparse.Namespace:
    """Setup the flags for the basic CLI"""

    parser = argparse.ArgumentParser(
        prog=__about__.__name__, description=__about__.__summary__
    )

    parser.add_argument(
        "--server", action="store_true", help="Run the Flask development server"
    )
    parser.add_argument(
        "--config", action="store_true", help="Check that the configuration is valid"
    )
    parser.add_argument(
        "--version", action="store_true", help="Show program version and exit"
    )
    parser.add_argument(
        "--index", action="store_true", help="Rebuild the source image index"
    )

    return parser.parse_args()


def main() -> int:
    """Main entrypoint for the CLI tooling"""
    args = get_args()

    logging.basicConfig(
        stream=sys.stderr,
        format="(%(asctime)s) %(levelname)s: %(message)s",
        datefmt="%I:%M:%S",
        level=logging.DEBUG,
    )

    if args.version:
        print(f"{__about__.__title__} {__about__.__version__}", file=sys.stderr)
        return 0

    if args.config:
        try:
            configuration.load()
        except Exception as err:  # pylint: disable=broad-except
            print(f"Configuration check failed: {err}", file=sys.stderr)
            return 1
        else:
            print("Configuration check successful!", file=sys.stderr)
            return 0

    if args.index:
        config = configuration.load()
        database.initialize(config)
        index.build(config)
        return 0

    if args.server:
        from kodak import application  # pylint: disable=import-outside-toplevel

        application.APPLICATION.run(host="127.0.0.1")
        return 0

    return 1


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kodak', 'kodak.database', 'kodak.resources', 'tests']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8.3.2,<9.0.0',
 'flask-restful>=0.3.8,<0.4.0',
 'flask>=1.1.2,<2.0.0',
 'peewee>=3.13.3,<4.0.0',
 'ruamel.yaml>=0.16.12,<0.17.0']

extras_require = \
{'deployment': ['gunicorn>=20.1.0,<21.0.0']}

entry_points = \
{'console_scripts': ['kodak = kodak.cli:main']}

setup_kwargs = {
    'name': 'kodak',
    'version': '0.1.0',
    'description': 'HTTP server for uploading images and generating thumbnails',
    'long_description': "# kodak\n\nWeb server for auto-generating banners, previews, thumbnails, and more from any directory.\nLightweight, simple, and designed for performance.\n\nDeveloped with [Poetry 1.0+](https://python-poetry.org/)\n\n## Goals\n\n- Support defining server-side manipulation specifications\n\n  ```\n  KODAK_MANIP_FOOBAR_CROP_VERTICAL=300\n  KODAK_MANIP_FOOBAR_SCALE_HORIZONTAL=1200\n  KODAK_MANIP_FOOBAR_SCALE_STRATEGY=absolute\n\n  KODAK_MANIP_FIZZBUZZ_NAME=black+white\n  KODAK_MANIP_FIZZBUZZ_BLACK_AND_WHITE=true\n  KODAK_MANIP_FIZZBUZZ_SCALE_HORIZONTAL=50\n  KODAK_MANIP_FIZZBUZZ_SCALE_STRATEGY=relative\n  ```\n\n- Support retrieving manipulated images based on server side configuration\n\n  ```\n  GET /image/<name>/foobar\n\n  GET /image/<name>/black+white\n  ```\n\n- Support optionally exposing full-resolution source images\n\n  ```\n  GET /image/<name>/original\n  ```\n\n- Support caching of generated image manipulations for reuse\n\n- Support [HTTP 410](https://httpstatuses.com/410) for indicating removed images and\n  manipulations\n\n- Support optional authentication with pre-generated access tokens\n\n- Support static file tree management for exposure via external web server (which is faster\n  than serving files with python)\n\n- Support automatic indexing of newly added image files\n\n- Support automatic indexing of removed image files\n\n- Support arbitrary source directory structure\n\n- Support Dockerized deployment\n\n- Support bare-metal deployment (via systemd)\n\n## Non-goals\n\n- Client-defined image manipulations through publicly exposed parameters\n\n  > Manipulating images is- in the grand scheme of things- pretty resource intensive. Exposing\n  > dynamic parameters that can be cycled through to generate hundreds or thousands of\n  > permutations for every known image on a server could be used to either consume the\n  > server's entire disk or server's entire CPU.\n\n- Upload functionality\n\n  > This application should be as simple as possible. Lots of people have implemented file\n  > upload systems, synchronizers, and managers way better than I have.\n\n- Robust and flexible access control\n\n  > See above. Complex authentication can be added using a reverse proxy or any one of several\n  > dozen options for 3rd party middleware. The provided authentication is supposed to be\n  > dead simple for people who absolutely need the server to be private but absolutely cannot\n  > implement something more complicated.\n\n- Pre-creation of image manipulations\n\n  > The goal of this program is just-in-time creation of the manipulated assets with\n  > aggressive caching; first load is slow, subsequent loads are fast. For this use case\n  > there's no sense creating or storing an asset until it's known to be needed.\n",
    'author': 'Ethan Paul',
    'author_email': '24588726+enpaul@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/enpaul/kodak/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
